#!/usr/bin/env python3
# I suspect that I haven't got to grips with the way phthon does things, but soppusely this will setup the path to allod for the sonnen batteri moduel to be in a separate location
# To me having to do this for testing seems a horrendous hack
import os, sys
import threading
script_path = os.path.realpath(os.path.dirname(__name__))
os.chdir(script_path)
sys.path.append("..")
from login import *
from pprint import pprint
from sonnenbatterie2.sonnenbatterie2 import SonnenBatterieV2
from sonnenbatterie2.const import *

def test_set_flow():
  sb = SonnenBatterieV2(SONNEN_IP, SONNEN_TOKEN)
  operating_mode_name = sb.get_operating_mode_name()
  current_flow = sb.get_status()["Pac_total_W"]
  print(f"current total flow is {current_flow}")
  print(f"Setting operating mode name to {SONNEN_OPERATING_MODE_MANUAL_NAME}")
  pprint(sb.set_operating_mode_by_name(SONNEN_OPERATING_MODE_MANUAL_NAME))
  e = threading.Event() 
  print("Waiting 5 seconds for things to settle down")
  e.wait(5)
  manual_flow=sb.get_status()["Pac_total_W"]
  print(f"Flow on manual is {manual_flow}")
  print("Setting a charge rate of 100") 
  # set to a charge rate of 100
  set_resp = sb.charge_battery(100)
  print(f"response to set charge is {set_resp}")
  print("Waiting 5 seconds for things to settle down")
  e.wait(5)
  manual_flow=sb.get_status()["Pac_total_W"]
  print(f"Flow with charge rate of 100 is {manual_flow}")
  print("Setting a discharge rate of 100") 
  # set to a charge rate of 100
  set_resp = sb.discharge_battery(100)
  print(f"response to set discharge is {set_resp}")
  print("Waiting 5 for things to settle down")
  e.wait(5)
  manual_flow=sb.get_status()["Pac_total_W"]
  print(f"Flow with discharge rate of 100 is {manual_flow}")
  print(f"Returning operating mode to origional mode of {operating_mode_name}")
  pprint(sb.set_operating_mode_by_name(operating_mode_name))

if __name__ == '__main__':
  test_set_flow()


#!/usr/bin/env python3
import asyncio
import os
import sys
import time

script_path = os.path.realpath(os.path.dirname(__name__))
os.chdir(script_path)
sys.path.append("..")

from sonnenbatterie2.sonnenbatterie2 import SonnenBatterieV2, AsyncSonnenBatterieV2

from pprint import pprint

from login import *

def main():
    print("\nTesting SonnenBatterieV2 - Sync API\n")
    sb2 = SonnenBatterieV2(SONNEN_IP, SONNEN_TOKEN)
    pprint(sb2.get_configurations())  # retrieve configuration overview
    pprint(sb2.get_battery_module_data())  # get battery module data
    pprint(sb2.get_inverter_data())  # retrieve inverter data
    pprint(sb2.get_latest_data())  # get latest date from sonnenbatterie
    pprint(sb2.get_powermeter_data())  # get data from power meters
    pprint(sb2.get_status())  # get overall status information
    pprint(sb2.get_io_data())  # get io status

async def async_main():
    print("\nTesting AsyncSonnenBatterieV2 - ASync API\n")
    sb2 = AsyncSonnenBatterieV2(SONNEN_IP, SONNEN_TOKEN)
    pprint(await sb2.get_configurations())  # retrieve configuration overview
    pprint(await sb2.get_battery_module_data())  # get battery module data
    pprint(await sb2.get_inverter_data())  # retrieve inverter data
    pprint(await sb2.get_latest_data())  # get latest date from sonnenbatterie
    pprint(await sb2.get_powermeter_data())  # get data from power meters
    pprint(await sb2.get_status())  # get overall status information
    pprint(await sb2.get_io_data())  # get io status
    await sb2.logout()

if __name__ == "__main__":
    start = time.time()
    main()
    print("--- %s seconds sync ---" % (time.time() - start))
    start = time.time()
    asyncio.run(async_main())
    print("--- %s seconds async ---" % (time.time() - start))


import json
import sys
import base64, binascii, hashlib, hmac, string

from sonnenbatterie2.sonnenbatterie2 import AsyncSonnenBatterieV2

sys.path.append("..")
import hashlib
import requests

import aiohttp

from sonnenbatterie2 import SonnenBatterieV2
from .const import *


# noinspection PyPep8Naming
class sonnenbatterie:
    # noinspection HttpUrlsUsage
    def __init__(self,username,password,ipaddress):
        self.username=username
        self.password=password
        self.ipaddress=ipaddress
        self.baseurl='http://'+self.ipaddress+'/api/'
        self.setpoint='v2/setpoint/'
        self._batteryLoginTimeout = DEFAULT_BATTERY_LOGIN_TIMEOUT 
        self._batteryConnectTimeout = DEFAULT_CONNECT_TO_BATTERY_TIMEOUT 
        self._batteryReadTimeout = DEFAULT_READ_FROM_BATTERY_TIMEOUT 
        self._batteryRequestTimeout = (self._batteryConnectTimeout, self._batteryReadTimeout)

        self.login()
        self.token = None
        self.sb2 = SonnenBatterieV2(ip_address=self.ipaddress, api_token=self.token)


    def login(self):
        password_sha512 = hashlib.sha512(self.password.encode('utf-8')).hexdigest()
        req_challenge=requests.get(self.baseurl+'challenge', timeout=self._batteryLoginTimeout)
        req_challenge.raise_for_status()

        salt=requests.get(self.baseurl+'salt/'+self.username, timeout=self._batteryLoginTimeout, allow_redirects=False)##returns a solt after battery got updated to a certatin version (>1.18??)
        saltResponseCode=salt.status_code

        




        challenge=req_challenge.json()
        if saltResponseCode!=200: #use old variant where no salt is availlable 
            #Old path (no salt): PBKDF2 over sha512(password) + challenge → hex
            response=hashlib.pbkdf2_hmac('sha512',password_sha512.encode('utf-8'),challenge.encode('utf-8'),7500,64).hex()
        else:#New path (with salt): PBKDF2 over password + salt → HMAC-SHA256(challenge) → hex.
            salt_payload=salt.json()
            salt_str = salt_payload["salt"]
            try:
                if len(salt_str) % 2 == 0 and all(c in string.hexdigits for c in salt_str):
                    salt_bytes = bytes.fromhex(salt_str)
                else:
                    raise ValueError
            except Exception:
                try:
                    salt_bytes = base64.b64decode(salt_str, validate=True)
                except Exception:
                    salt_bytes = salt_str.encode("utf-8")

            
            dk = hashlib.pbkdf2_hmac(
                "sha512",
                self.password.encode("utf-8"),  # raw password (not pre-hashed)
                salt_bytes,
                7500,
                dklen=64
            )
            response = hmac.new(dk, challenge.encode("utf-8"), hashlib.sha256).hexdigest()

        
        getsession=requests.post(self.baseurl+'session',{"user":self.username,"challenge":challenge,"response":response}, timeout=self._batteryLoginTimeout)
        getsession.raise_for_status()
        token=getsession.json()['authentication_token']
        self.token=token

    def set_login_timeout(self, timeout:int = 120):
        self._batteryLoginTimeout = timeout
    
    def get_login_timeout(self) -> int:
        return self._batteryLoginTimeout
    
    def set_request_connect_timeout(self, timeout:int = 60):
        self._batteryConnectTimeout = timeout
        self._batteryRequestTimeout = (self._batteryConnectTimeout, self._batteryReadTimeout)

    def get_request_connect_timeout(self) -> int:
        return self._batteryConnectTimeout
    
    def set_request_read_timeout(self, timeout:int = 60):
        self._batteryReadTimeout = timeout
        self._batteryRequestTimeout = (self._batteryConnectTimeout, self._batteryReadTimeout)

    def get_request_read_timeout(self) -> int:
        return self._batteryReadTimeout
    
    def _get(self,what,isretry=False):
        # This is a synchronous call, you may need to wrap it in a thread or something for asynchronous operation        
        url = self.baseurl+what
        response=requests.get(url,
            headers={'Auth-Token': self.token}, timeout=self._batteryRequestTimeout
        )
        if not isretry and response.status_code==401:
            self.login()
            return self._get(what,True)
        if response.status_code != 200:
            response.raise_for_status()

        return response.json()    

    def _put(self, what, payload, isretry=False):
        # This is a synchronous call, you may need to wrap it in a thread or something for asynchronous operation
        url = self.baseurl+what
        response=requests.put(url,
            headers={'Auth-Token': self.token,'Content-Type': 'application/json'} , json=payload, timeout=self._batteryRequestTimeout
        )
        if not isretry and response.status_code==401:
            self.login()
            return self._put(what, payload,True)
        if response.status_code != 200:
            response.raise_for_status()
        return response.json()

    def _post(self, what, isretry=False):
        # This is a synchronous call, you may need to wrap it in a thread or something for asynchronous operation
        url = self.baseurl+what
        response=requests.post(url,
            headers={'Auth-Token': self.token,'Content-Type': 'application/json'}, timeout=self._batteryRequestTimeout
        )
        if not isretry and response.status_code==401:
            self.login()
            return self._post(what, True)
        if response.status_code != 200:
            response.raise_for_status()
        return response
    
    # these are special purpose endpoints, there is no associated data that I'm aware of
    # while I don't have details I belive this is probabaly only useful in manual more
    # and it's probabaly possible to extact the actuall flow rate in operation  
    # looking at the status.state_battery_inout value
    # irritatingly there is no mechanism in the API to do a single set to you have to work out if
    # the direction of the flow and then call the appropriate API 
    # more general purpose endpoints

    # API v1 calls
    def get_powermeter(self):
        return self._get(SONNEN_API_PATH_POWER_METER)
        
    def get_batterysystem(self):
        return self._get(SONNEN_API_PATH_BATTERY_SYSTEM)
        
    def get_inverter(self):
        return self._get(SONNEN_API_PATH_INVERTER)
    
    def get_systemdata(self):
        return self._get(SONNEN_API_PATH_SYSTEM_DATA)

    def get_status(self):
        return self._get(SONNEN_API_PATH_STATUS)
        
    def get_battery(self):
        return self._get(SONNEN_API_PATH_BATTERY)

    def get_api_configuration(self):
        return self._get(SONNEN_API_PATH_API_CONFIGURATION)

    def get_commissioning_settings(self):
        return self._get(SONNEN_API_COMMISSIONING_SETTINGS)


    # API v2 calls
    def set_configuration(self, name, value):
        return self.sb2.set_config_item(name, value)

    def get_latest_data(self):
        return self.sb2.get_latest_data()

    def get_configurations(self):
        return self.sb2.get_configurations()

    # deprecate with 0.7.0 -> get_config_item
    def get_configuration(self, name):
        return self.sb2.get_config_item(name)

    def get_config_item(self, name):
        return self.sb2.get_config_item(name)
    

    # these have special handling in some form, for example converting a mode as a number into a string
    def get_current_charge_level(self):
        return self.get_latest_data().get(SONNEN_LATEST_DATA_CHARGE_LEVEL)

    def get_operating_mode(self) -> int:
        return int(self.get_config_item(SONNEN_CONFIGURATION_OPERATING_MODE)[SONNEN_CONFIGURATION_OPERATING_MODE])
    
    def get_operating_mode_name(self):
        operating_mode = self.get_operating_mode()
        return SONNEN_OPERATING_MODES_TO_OPERATING_MODE_NAMES.get(operating_mode)
    
    def set_operating_mode(self, operating_mode) -> int:
        return int(self.set_configuration(SONNEN_CONFIGURATION_OPERATING_MODE, operating_mode))
    
    def set_operating_mode_by_name(self, operating_mode_name) -> int:
        return self.set_operating_mode(SONNEN_OPERATING_MODE_NAMES_TO_OPERATING_MODES.get(operating_mode_name))
    
    def get_battery_reserve(self):
        return self.get_config_item(SONNEN_CONFIGURATION_BACKUP_RESERVE)[SONNEN_CONFIGURATION_BACKUP_RESERVE]
    
    def set_battery_reserve(self, reserve=5):
        reserve = int(reserve)
        if (reserve < 0) or (reserve > 100):
            raise Exception(f"Reserve must be between 0 and 100, you specified {reserve}")
        return self.set_configuration(SONNEN_CONFIGURATION_BACKUP_RESERVE, reserve)
    
    # set the reserve to the current battery level adjusted by the offset if provided
    # (a positive offset means that the reserve will be set to more than the current level
    # a negative offser means less than the current level)
    # If the new reserve is less than the minimum reserve then use the minimum reserve
    # the reserve will be tested to ensure it's >= 0 or <= 100
    def set_battery_reserve_relative_to_current_charge(self, offset=0, minimum_reserve=0):
        current_level = self.get_current_charge_level()
        target_level = current_level +offset
        if target_level <  minimum_reserve:
            target_level = minimum_reserve
        if target_level < 0:
            target_level = 0
        elif target_level > 100:
            target_level = 100
        return self.set_battery_reserve(target_level)


""" AsyncSonnenbatterie """

class AsyncSonnenBatterie:
    # noinspection HttpUrlsUsage
    def __init__(self, username, password, ipaddress):
        self.username = username
        self.password = password
        self.ipaddress = ipaddress

        self.baseurl = 'http://' + self.ipaddress + '/api/'

        self._timeout_total = DEFAULT_BATTERY_LOGIN_TIMEOUT
        self._timeout_connect = DEFAULT_CONNECT_TO_BATTERY_TIMEOUT
        self._timeout_read = DEFAULT_READ_FROM_BATTERY_TIMEOUT

        self._timeout = self._set_timeouts()

        self.token = None
        self.sb2 = None

        self._session = None

    def _set_timeouts(self) -> aiohttp.ClientTimeout:
        return aiohttp.ClientTimeout(
            total=self._timeout_total,
            connect=self._timeout_connect,
            sock_connect=self._timeout_connect,
            sock_read=self._timeout_read,
        )

    async def logout(self):
        if self.sb2:
            await self.sb2.logout()
            self.sb2 = None

        if self._session:
            await self._session.close()
            self._session = None

    async def login(self):
        if self._session is None:
            self._session = aiohttp.ClientSession()

        pw_sha512 = hashlib.sha512(self.password.encode('utf-8')).hexdigest()
        req_challenge = await self._session.get(
            self.baseurl+'challenge',
            timeout=self._timeout,
        )
        req_challenge.raise_for_status()
        challenge = await req_challenge.json()

        salt =  await self._session.get(self.baseurl+'salt/'+self.username, timeout=self._timeout, allow_redirects=False)##returns a solt after battery got updated to a certatin version (>1.18??)
        saltResponseCode = salt.status




        if saltResponseCode!=200: #use old variant where no salt is availlable 
            #Old path (no salt): PBKDF2 over sha512(password) + challenge → hex
            response=hashlib.pbkdf2_hmac('sha512',pw_sha512.encode('utf-8'),challenge.encode('utf-8'),7500,64).hex()
            session = await self._session.post(
                url = self.baseurl+'session',
                data = {"user":self.username,"challenge":challenge,"response":response},
                timeout=self._timeout,
            )
            session.raise_for_status()
            token = await session.json()
            self.token = token['authentication_token']
        else:#New path (with salt): PBKDF2 over password + salt → HMAC-SHA256(challenge) → hex.
          await self.create_session_token_new()




        # Inisitalite async API v2
        if self.sb2 is None:
            self.sb2 = AsyncSonnenBatterieV2(ip_address=self.ipaddress, api_token=self.token)


    async def create_session_token_new(self):
        # Step 1: challenge
        async with self._session.get(f"{self.baseurl}challenge") as r:
            r.raise_for_status()
            try:
                challenge = await r.json()
                if isinstance(challenge, dict):
                    challenge = challenge.get("challenge") or next(iter(challenge.values()))
            except Exception:
                challenge = (await r.text()).strip()

        # Step 2: salt
        async with self._session.get(f"{self.baseurl}salt/{self.username}") as r:
            r.raise_for_status()
            salt = (await r.json())["salt"]

        # Step 3: derive key
        pw_sha512_hex = hashlib.sha512(self.password.encode("utf-8")).hexdigest()
        pw_bytes = pw_sha512_hex.encode("utf-8")
        dk = hashlib.pbkdf2_hmac("sha512", pw_bytes, salt.encode("utf-8"), 7500, dklen=64)
        derived_hex = dk.hex()

        # Step 4: loop: POST /session, maybe retry with new_challenge
        for _ in range(2):
            key = derived_hex.encode("utf-8")
            response = hmac.new(key, challenge.encode("utf-8"), hashlib.sha256).hexdigest()
            payload = {"user": self.username, "challenge": challenge, "response": response}

            async with self._session.post(f"{self.baseurl}session", json=payload) as r:
                txt = await r.text()
                if r.status >= 400:
                    raise RuntimeError(
                        f"Login failed with HTTP {r.status}\n"
                        f"URL: {r.url}\n"
                        f"Payload: {json.dumps(payload)}\n"
                        f"challenge: {challenge}\n"
                        f"salt: {salt}\n"
                        f"pw_sha_hex: {pw_sha512_hex}\n"
                        f"derived_hex: {derived_hex}\n"
                        f"response_hex: {response}\n"
                        f"Server reply: {txt}"
                    )
                r.raise_for_status()
                data = await r.json()

            if "new_challenge" in data and data["new_challenge"]:
                challenge = data["new_challenge"]
                continue

            if "authentication_token" in data and data["authentication_token"]:
                 self.token = data["authentication_token"]

            raise RuntimeError(f"Unexpected /session response: {json.dumps(data)}")

        raise RuntimeError("Failed to obtain authentication_token")




    """ Base functions """


    async def _get(self, what, isretry=False) -> json:
        if self.token is None:
            await self.login()

        url = self.baseurl + what
        response = await self._session.get(
            url = url,
            headers={'Auth-Token': self.token},
            timeout=self._timeout,
        )

        if not isretry and response.status == 401:
            await self.login()
            return await self._get(what, True)

        if response.status != 200:
            response.raise_for_status()

        return await response.json()


    async def _post(self, what, isretry=False) -> json:
        if self.token is None:
            await self.login()

        url = self.baseurl + what
        response = await self._session.post(
            url = url,
            headers = {'Auth-Token': self.token, 'Content-Type': 'application/json'},
            timeout = self._timeout,
        )

        if not isretry and response.status == 401:
            await self.login()
            return await self._post(what, True)
        if response.status != 200:
            response.raise_for_status()

        return await response.json()


    async def _put(self, what, payload, isretry=False) -> json:
        if self.token is None:
            await self.login()

        url = self.baseurl + what
        response = await self._session.put(
            url = url,
            headers = {'Auth-Token': self.token, 'Content-Type': 'application/json'},
            json = payload,
            timeout = self._timeout,
        )

        if not isretry and response.status == 401:
            await self.login()
            return await self._put(what, True)

        if response.status != 200:
            response.raise_for_status()

        return await response.json()


    """ API functions """


    async def get_powermeter(self) -> json:
        result = await self._get(SONNEN_API_PATH_POWER_METER)
        return result

    async def get_batterysystem(self) -> json:
        return await self._get(SONNEN_API_PATH_BATTERY_SYSTEM)

    async def get_inverter(self) -> json:
        return await self._get(SONNEN_API_PATH_INVERTER)

    async def get_systemdata(self) -> json:
        return await self._get(SONNEN_API_PATH_SYSTEM_DATA)

    async def get_status(self) -> json:
        return await self._get(SONNEN_API_PATH_STATUS)

    async def get_battery(self) -> json:
        return await self._get(SONNEN_API_PATH_BATTERY)

    async def get_api_configuration(self) -> json:
        return await self._get(SONNEN_API_PATH_API_CONFIGURATION)

    async def get_commissioning_settings(self):
        return await self._get(SONNEN_API_COMMISSIONING_SETTINGS)


    """ API v2 calls """
    async def set_configuration(self, name, value) -> json:
        if self.sb2 is None:
            await self.login()
        return await self.sb2.set_config_item(name, value)

    async def get_latest_data(self) -> json:
        if self.sb2 is None:
            await self.login()
        return await self.sb2.get_latest_data()

    async def get_configurations(self) -> json:
        if self.sb2 is None:
            await self.login()
        return await self.sb2.get_configurations()

    async def get_config_item(self, name) -> json:
        if self.sb2 is None:
            await self.login()
        return await self.sb2.get_config_item(name)


    """ Special functions """
    # GET
    async def get_current_charge_level(self) -> int:
        result = await self.get_latest_data()
        return result.get(SONNEN_LATEST_DATA_CHARGE_LEVEL)

    async def get_operating_mode(self) -> int:
        result = await self.get_config_item(SONNEN_CONFIGURATION_OPERATING_MODE)
        return int(result[SONNEN_CONFIGURATION_OPERATING_MODE])

    async def get_operating_mode_name(self) -> str:
        opmode = await self.get_operating_mode()
        return SONNEN_OPERATING_MODES_TO_OPERATING_MODE_NAMES.get(opmode)

    async def get_battery_reserve(self) -> int:
        result = await self.get_config_item(SONNEN_CONFIGURATION_BACKUP_RESERVE)
        return int(result[SONNEN_CONFIGURATION_BACKUP_RESERVE])

    #SET
    async def set_operating_mode(self, operating_mode) -> int:
        result = await self.set_configuration(
            SONNEN_CONFIGURATION_OPERATING_MODE,
            operating_mode
        )
        return int(result[SONNEN_CONFIGURATION_OPERATING_MODE])

    async def set_operating_mode_by_name(self, operating_mode_name) -> int:
        result = await self.set_configuration(
            SONNEN_CONFIGURATION_OPERATING_MODE,
            SONNEN_OPERATING_MODE_NAMES_TO_OPERATING_MODES.get(operating_mode_name)
        )
        return int(result[SONNEN_CONFIGURATION_OPERATING_MODE])

    async def set_battery_reserve(self, battery_reserve) -> json:
        reserve = int(battery_reserve)
        if (reserve < 0) or (reserve > 100):
            raise Exception(f"Reserve must be between 0 and 100, you spcified {reserve}")
        return await self.set_configuration(SONNEN_CONFIGURATION_BACKUP_RESERVE, reserve)

    # noinspection DuplicatedCode
    async def set_battery_reserve_relative_to_current_charge(self, offset=0, min_res=0) -> json:
        current_level = await self.get_current_charge_level()
        target_level = current_level + offset
        if target_level < min_res:
            target_level = min_res
        if target_level > 100:
            target_level = 100
        if target_level < 0:
            target_level = 0
        return await self.set_battery_reserve(target_level)



import json

import aiohttp
import requests

from .const import (
    DEFAULT_CONNECT_TO_BATTERY_TIMEOUT,
    DEFAULT_READ_FROM_BATTERY_TIMEOUT,
    SONNEN_CONFIGURATION_TOU_SCHEDULE, DEFAULT_BATTERY_LOGIN_TIMEOUT, SONNEN_LATEST_DATA_CHARGE_LEVEL,
    SONNEN_CONFIGURATION_OPERATING_MODE, SONNEN_OPERATING_MODES_TO_OPERATING_MODE_NAMES,
    SONNEN_CONFIGURATION_BACKUP_RESERVE, SONNEN_OPERATING_MODE_NAMES_TO_OPERATING_MODES
)
from timeofuse import TimeofUseSchedule
from timeofuse.timeofuse import create_time_of_use_schedule_from_json


class SonnenBatterieV2:
    def __init__(self, ip_address, api_token):
        self.api_token = api_token
        self.ip_address = ip_address
        # noinspection HttpUrlsUsage
        self.base_url = f'http://{ip_address}/api/v2/'
        self._batteryConnectTimeout = DEFAULT_CONNECT_TO_BATTERY_TIMEOUT
        self._batteryReadTimeout = DEFAULT_READ_FROM_BATTERY_TIMEOUT
        self._batteryRequestTimeout = (self._batteryConnectTimeout, self._batteryReadTimeout)

    def set_request_connect_timeout(self, timeout:int = 60):
        self._batteryConnectTimeout = timeout
        self._batteryRequestTimeout = (self._batteryConnectTimeout, self._batteryRequestTimeout)

    def get_request_connect_timeout(self) -> int:
        return self._batteryConnectTimeout

    def set_request_read_timeout(self, timeout:int = 120):
        self._batteryReadTimeout = timeout
        self._batteryRequestTimeout = (self._batteryReadTimeout, self._batteryRequestTimeout)

    def get_request_read_timeout(self) -> int:
        return self._batteryReadTimeout

    def _get(self, what, isretry=False):
        url = self.base_url + what
        response = requests.get(
            url,
            headers={'Auth-Token': self.api_token},
            timeout=self._batteryRequestTimeout,
        )
        if not isretry and response.status_code == 401:
            return self._get(what, isretry=True)

        if response.status_code != 200:
            response.raise_for_status()

        return response.json()

    def _put(self, what, payload, isretry=False):
        url = self.base_url + what
        response = requests.put(
            url,
            headers={'Auth-Token': self.api_token, 'Content-Type': 'application/json'},
            json=payload,
            timeout=self._batteryRequestTimeout,
        )
        if not isretry and response.status_code == 401:
            return self._put(what, payload, isretry=True)

        if response.status_code != 200:
            response.raise_for_status()

        return response.json()

    def _post(self, what, isretry=False):
        url = self.base_url + what
        response = requests.post(
            url,
            headers={'Auth-Token': self.api_token, 'Content-Type': 'application/json'},
            timeout=self._batteryRequestTimeout,
        )
        if not isretry and response.status_code == 401:
            return self._post(what, isretry=True)

        if response.status_code != 200:
            response.raise_for_status()

        return response.json()

    # GET API
    def get_config_item(self, item):
        return self._get(f"configurations/{item}")

    def get_configurations(self):
        return self._get("configurations")

    def get_battery_module_data(self):
        return self._get("battery")

    def get_inverter_data(self):
        return self._get("inverter")

    def get_latest_data(self):
        return self._get("latestdata")

    def get_powermeter_data(self):
        return self._get("powermeter")

    def get_status(self):
        return self._get("status")

    # deprecate with 0.7.x -> get_status
    def get_status_data(self):
        return self._get("status")

    def get_io_data(self):
        return self._get("io")

    # POST API
    def set_manual_flow(self, direction, watts) -> bool:
        response = self._post(f"setpoint/{direction}/{watts}")
        return response

    def charge_battery(self, watts:int):
        return self.set_manual_flow("charge", watts)

    def discharge_battery(self, watts:int):
        return self.set_manual_flow("discharge", watts)

    # PUT API
    """ set item in configurations and return value as reported back
        by the SonnenBatterieV2 API
    """
    def set_config_item(self, item:str, value:str|int) -> str:
        payload = {str(item): str(value)}
        return self._put(f"configurations", payload)[item]

    # API Helpers
    def get_current_charge_level(self):
        return self.get_latest_data().get(SONNEN_LATEST_DATA_CHARGE_LEVEL)

    def get_operating_mode(self) -> int:
        return int(self.get_config_item(SONNEN_CONFIGURATION_OPERATING_MODE)[SONNEN_CONFIGURATION_OPERATING_MODE])

    def get_operating_mode_name(self):
        operating_mode = self.get_operating_mode()
        return SONNEN_OPERATING_MODES_TO_OPERATING_MODE_NAMES.get(operating_mode)

    def set_operating_mode(self, operating_mode) -> int:
        return int(self.set_config_item(SONNEN_CONFIGURATION_OPERATING_MODE, operating_mode))

    def set_operating_mode_by_name(self, operating_mode_name) -> int:
        return int(self.set_operating_mode(SONNEN_OPERATING_MODE_NAMES_TO_OPERATING_MODES.get(operating_mode_name)))

    def get_battery_reserve(self):
        return self.get_config_item(SONNEN_CONFIGURATION_BACKUP_RESERVE)[SONNEN_CONFIGURATION_BACKUP_RESERVE]

    def set_battery_reserve(self, reserve=5):
        reserve = int(reserve)
        if (reserve < 0) or (reserve > 100):
            raise Exception(f"Reserve must be between 0 and 100, you specified {reserve}")
        return self.set_config_item(SONNEN_CONFIGURATION_BACKUP_RESERVE, reserve)

    # set the reserve to the current battery level adjusted by the offset if provided
    # (a positive offset means that the reserve will be set to more than the current level
    # a negative offser means less than the current level)
    # If the new reserve is less than the minimum reserve then use the minimum reserve
    # the reserve will be tested to ensure it's >= 0 or <= 100
    def set_battery_reserve_relative_to_current_charge(self, offset=0, minimum_reserve=0):
        current_level = self.get_current_charge_level()
        target_level = current_level + offset
        if target_level < minimum_reserve:
            target_level = minimum_reserve
        if target_level < 0:
            target_level = 0
        elif target_level > 100:
            target_level = 100
        return self.set_battery_reserve(target_level)

    def get_tou_schedule_string(self) -> str:
        return self.get_config_item(SONNEN_CONFIGURATION_TOU_SCHEDULE)[SONNEN_CONFIGURATION_TOU_SCHEDULE]

    def get_tou_schedule_json(self) -> json:
        return json.loads(self.get_tou_schedule_string())

    def get_tou_schedule_object(self) -> TimeofUseSchedule:
        return create_time_of_use_schedule_from_json(self.get_tou_schedule_json())

    def set_tou_schedule_string(self, schedule:str):
        return self.set_config_item(SONNEN_CONFIGURATION_TOU_SCHEDULE, schedule)

    def set_tou_schedule_json(self, schedule):
        # We need to convert object to valid json string
        return self.set_tou_schedule_string(json.dumps(schedule))

    def clear_tou_schedule_string(self):
        return self.set_tou_schedule_string("[]")

    def clear_tou_schedule_json(self):
        return self.set_tou_schedule_string("[]")


""" Async API calls """


class AsyncSonnenBatterieV2:
    def __init__(self, ip_address, api_token):
        self._ip_address = ip_address
        self._api_token = api_token

        # noinspection HttpUrlsUsage
        self.base_url = f"http://{self._ip_address}/api/v2/"

        self._timeout_total = DEFAULT_BATTERY_LOGIN_TIMEOUT
        self._timeout_connect = DEFAULT_CONNECT_TO_BATTERY_TIMEOUT
        self._timeout_read = DEFAULT_READ_FROM_BATTERY_TIMEOUT
        self._timeout = self._set_timeouts()

        self._session = None

    async def logout(self):
        if self._session:
            await self._session.close()
            self._session = None

    def _set_timeouts(self) -> aiohttp.ClientTimeout:
        return aiohttp.ClientTimeout(
            total=self._timeout_total,
            connect=self._timeout_connect,
            sock_connect=self._timeout_connect,
            sock_read=self._timeout_read,
        )


    """ Base functions """


    async def _get(self, what, isretry=False) -> json:
        if self._session is None:
            self._session = aiohttp.ClientSession()

        url = self.base_url + what
        response = await self._session.get(
            url=url,
            headers={'Auth-Token': self._api_token},
            timeout=self._timeout,
        )

        if not isretry and response.status == 401:
            return await self._get(what, isretry=True)
        if response.status != 200:
            response.raise_for_status()

        return await response.json()


    async def _put(self, what, payload, isretry=False):
        if self._session is None:
            self._session = aiohttp.ClientSession()

        url = self.base_url + what
        response = await self._session.put(
            url=url,
            headers={'Auth-Token': self._api_token},
            json=payload,
            timeout=self._timeout,
        )


        if not isretry and response.status == 401:
            return await self._put(what, payload, isretry=True)
        if response.status != 200:
            response.raise_for_status()

        return await response.json()


    async def _post(self, what, isretry=False) -> json:
        if self._session is None:
            self._session = aiohttp.ClientSession()

        url = self.base_url + what
        response = await self._session.post(
            url=url,
            headers={'Auth-Token': self._api_token},
            timeout=self._timeout,
        )

        if not isretry and response.status == 401:
            return await self._post(what, isretry=True)
        if response.status != 200:
            response.raise_for_status()

        return await response.json()


    """ API function """

    # GET
    async def get_config_item(self, item:str) -> json:
        return await self._get(f"configurations/{item}")

    async def get_configurations(self) -> json:
        return await self._get("configurations")

    async def get_battery_module_data(self) -> json:
        return await self._get("battery")

    async def get_inverter_data(self) -> json:
        return await self._get("inverter")

    async def get_latest_data(self) -> json:
        return await self._get("latestdata")

    async def get_powermeter_data(self) -> json:
        return await self._get("powermeter")

    async def get_status(self) -> json:
        return await self._get("status")

    async def get_status_data(self) -> json:
        return await self._get("status")

    async def get_io_data(self) -> json:
        return await self._get("io")

    # POST
    async def set_manual_flow(self, direction, watts) -> json:
        return await self._post(f"setpoint/{direction}/{watts}")

    async def charge_battery(self, watts:int) -> json:
        return await self.set_manual_flow("charge", watts)

    async def discharge_battery(self, watts:int) -> json:
        return await self.set_manual_flow("discharge", watts)

    # PUT
    async def set_config_item(self, item:str, value:str|int) -> json:
        payload = {str(item): str(value)}
        return await self._put(f"configurations", payload)

    # API Helpers
    async def get_current_charge_level(self) -> int:
        result = await self.get_latest_data()
        return result.get(SONNEN_LATEST_DATA_CHARGE_LEVEL)

    async def get_operating_mode(self) -> int:
        result = await self.get_config_item(SONNEN_CONFIGURATION_OPERATING_MODE)
        return int(result[SONNEN_CONFIGURATION_OPERATING_MODE])

    async def get_operating_mode_name(self) -> str:
        opmode = await self.get_operating_mode()
        return SONNEN_OPERATING_MODES_TO_OPERATING_MODE_NAMES.get(opmode)

    async def get_battery_reserve(self) -> int:
        result = await self.get_config_item(SONNEN_CONFIGURATION_BACKUP_RESERVE)
        return int(result[SONNEN_CONFIGURATION_BACKUP_RESERVE])

    async def set_operating_mode(self, operating_mode) -> int:
        result = await self.set_config_item(
            SONNEN_CONFIGURATION_OPERATING_MODE,
            operating_mode
        )
        return int(result[SONNEN_CONFIGURATION_OPERATING_MODE])

    async def set_operating_mode_by_name(self, operating_mode_name) -> int:
        result = await self.set_config_item(
            SONNEN_CONFIGURATION_OPERATING_MODE,
            SONNEN_OPERATING_MODE_NAMES_TO_OPERATING_MODES.get(operating_mode_name)
        )
        return int(result[SONNEN_CONFIGURATION_OPERATING_MODE])

    async def set_battery_reserve(self, battery_reserve) -> json:
        reserve = int(battery_reserve)
        if (reserve < 0) or (reserve > 100):
            raise Exception(f"Reserve must be between 0 and 100, you spcified {reserve}")
        return await self.set_config_item(SONNEN_CONFIGURATION_BACKUP_RESERVE, reserve)

    # noinspection DuplicatedCode
    async def set_battery_reserve_relative_to_current_charge(self, offset=0, min_res=0) -> json:
        current_level = await self.get_current_charge_level()
        target_level = current_level + offset
        if target_level < min_res:
            target_level = min_res
        if target_level > 100:
            target_level = 100
        if target_level < 0:
            target_level = 0
        return await self.set_battery_reserve(target_level)

    async def get_tou_schedule_string(self) -> str:
        return (await self.get_config_item(SONNEN_CONFIGURATION_TOU_SCHEDULE))[SONNEN_CONFIGURATION_TOU_SCHEDULE]

    async def get_tou_schedule_json(self) -> json:
        return json.loads(await self.get_tou_schedule_string())

    async def get_tou_schedule_object(self) -> json:
        return create_time_of_use_schedule_from_json(await self.get_tou_schedule_json())

    async def set_tou_schedule_string(self, schedule:str) -> json:
        return await self.set_config_item(SONNEN_CONFIGURATION_TOU_SCHEDULE, schedule)

    async def set_tou_schedule_json(self, schedule:str|list) -> json:
        # We need to convert object to valid json string
        return await self.set_tou_schedule_string(json.dumps(schedule))

    async def clear_tou_schedule_string(self) -> json:
        return await self.set_tou_schedule_string("[]")

    async def clear_tou_schedule_json(self) -> json:
        return await self.set_tou_schedule_string("[]")



#!/usr/bin/env python3
# I suspect that I haven't got to grips with the way phthon does things, but soppusely this will setup the path to allod for the sonnen batteri moduel to be in a separate location
# To me having to do this for testing seems a horrendous hack
import os, sys
script_path = os.path.realpath(os.path.dirname(__name__))
os.chdir(script_path)
sys.path.append("..")
from login import *
from pprint import pprint
from sonnenbatterie.sonnenbatterie import sonnenbatterie
if __name__ == '__main__':
  sb = sonnenbatterie(SONNEN_USERNAME, SONNEN_PASSWORD, SONNEN_IP)
  battery_reserve = sb.get_battery_reserve()
  current_charge = sb.get_current_charge_level()
  print("\nCurrent Configuration")
  pprint(sb.get_configurations())
  print("\nCurrent charge level")
  pprint(sb.get_current_charge_level())
  print("\nCurrent battery reserve")
  pprint(sb.get_battery_reserve())
  print("\nSetting absolute reserve to 7")
  pprint(sb.set_battery_reserve(7))
  print("\nUpdated battery reserve")
  pprint(sb.get_battery_reserve())
  print("\nSet relative limit (no offset)")
  pprint(sb.set_battery_reserve_relative_to_current_charge())
  print("\nBattery reserve with no offset against charge of "+str(current_charge))
  pprint(sb.get_battery_reserve())
  print("\nSet relative limit (offset of -5)")
  pprint(sb.set_battery_reserve_relative_to_current_charge(-5))
  print("\nBattery reserve with offset of -5 against charge of "+str(current_charge))
  pprint(sb.get_battery_reserve())
  print("\nSet relative limit (offset of 0 and minimum of 20)")
  pprint(sb.set_battery_reserve_relative_to_current_charge(-5, 20))
  print("\nBattery reserve with offset of 0 against minimum charge of 20")
  pprint(sb.get_battery_reserve())
  print("\nSet relative limit (offset of 0 and minimum of 80)")
  pprint(sb.set_battery_reserve_relative_to_current_charge(0, 80))
  print("\nBattery reserve with offset of 0 against minimum charge of 80")
  pprint(sb.get_battery_reserve())
  print("\nGoing to battery reserve of 10")
  pprint(sb.set_battery_reserve(10))
  print("\nReset battery reserve")
  pprint(sb.set_battery_reserve(battery_reserve))
  print("\nVerify battery reserve")
  pprint(sb.get_battery_reserve())



"""Process the time of use schedule stuff"""
from datetime import time, datetime
from typing import List


ATTR_TOU_START="start"
ATTR_TOU_STOP="stop"
ATTR_TOU_MAX_POWER="threshold_p_max"
TIME_FORMAT="%H:%M"

MIDNIGHT=time()

# sru_241217: moved to outside of class(es)
def create_time_of_use_entry(start_hour=23, start_min=30, stop_hour=5, stop_min=30, max_power=20000):
    start_time = time(hour=start_hour, minute=start_min)
    stop_time = time(hour=stop_hour, minute=stop_min)
    return TimeofUse(start_time, stop_time, int(max_power))

def create_time_of_use_schedule_from_json(json_schedule) :
    tous = TimeofUseSchedule()
    for entry in json_schedule:
        tou_entry = create_time_of_use_from_tou(entry)
        tous.add_entry(tou_entry)
    return tous

def create_time_of_use_from_tou(tou):
    # parse it out
    start_time = datetime.strptime(tou.get(ATTR_TOU_START), TIME_FORMAT).time()
    stop_time = datetime.strptime(tou.get(ATTR_TOU_STOP), TIME_FORMAT).time()
    max_power= int(tou.get(ATTR_TOU_MAX_POWER))
    # build the resulting object
    return TimeofUse(start_time, stop_time, max_power)


def _get_time_as_string(timeobj:time) -> str:
    return timeobj.strftime(TIME_FORMAT)


class TimeofUse:
    # sru_241217: We make this only accept named parameters
    def __init__(self, start_time:time, stop_time:time, max_power:int=20000):
        self.start_time = start_time
        # for comparisson reasons we can't have the stop time be midnight as thaty's actually the earliest possible time
        # so if it is provided as midnight make it the latest possible time
        if stop_time == MIDNIGHT:
            stop_time = time.max
        self.stop_time = stop_time
        self.max_power = max_power

    def __eq__(self, other) -> bool :
        if not isinstance(other, type(self)):
            return False
        return (self.start_time == other.start_time) and (self.stop_time == other.stop_time) and (self.max_power == other.max_power) 

    def __hash__(self) -> int:
        return hash(self.start_time) ^ hash(self.stop_time) ^ hash(self.max_power)

    def get_as_tou(self):
        start_string = self.get_start_time_as_string()
        tmp_stop_time = self.stop_time
        # previously we had to munge midnight as an stop time to 23:59:59.999999
        # if that was done now undo it
        if tmp_stop_time == time.max:
            tmp_stop_time = MIDNIGHT

        stop_string=tmp_stop_time.strftime(TIME_FORMAT)
        # max_power_string = str(self.max_power) ### sru_241217: we DON'T want this as string!
        return {ATTR_TOU_START:start_string, ATTR_TOU_STOP:stop_string, ATTR_TOU_MAX_POWER:self.max_power}
    
    def get_as_string(self) -> str:
        resp = "Start " + self.get_start_time_as_string() + ", End " +self.get_stop_time_as_string() +", Max allowable power " + str(self.get_max_power())
        return resp 

    def get_start_time_as_string(self) -> str:
        return _get_time_as_string(self.start_time)
    
    def get_stop_time_as_string(self) -> str:
        tmp_stop_time = self.stop_time
        # previously we had to munge midnight as an stop time to 23:59:59.999999
        # if that was done now undo it
        if tmp_stop_time == time.max:
            tmp_stop_time = MIDNIGHT
        return _get_time_as_string(tmp_stop_time)

    def get_max_power(self) -> int:
        return self.max_power

    def is_overlapping(self, other):
        # is our start time within the others time window ?
        if (self.start_time >= other.start_time) and (self.start_time <= other.stop_time):
            return True

        # is our end time within the others time window ?
        if (self.stop_time > other.start_time) and (self.stop_time <= other.stop_time):
            return True

        # is it's start time within the out time window ?
        if (other.start_time >= self.start_time) and (other.start_time < self.stop_time):
            return True

        # is it's end time within the out time window ?
        if (other.stop_time >= self.start_time) and (other.stop_time <= self.stop_time):
            return True

        # no overlap
        return False

class TimeofUseSchedule:
    def __init__(self):
        self._schedule_entries = []

    # adds the entry ensuring that it does not overlap with an existing entry
    def _add_entry(self, entry):
        if entry.stop_time < entry.start_time:
            raise Exception("End time cannot be before start time")
        for i in self._schedule_entries:
            if i.is_overlapping(entry):
                raise Exception("Unable to add entry, overlaps with exisiting entry")
        self._schedule_entries.append(entry)
        # maintains this as a sotred list based on the start time, this lets us compare properly
        self._schedule_entries = sorted(self._schedule_entries, key=lambda l_entry: l_entry.start_time)

    # Add an entry, if it spans midnight split it into a before midnight and after midnight section
    # Note that this IS NOT reversed on retrieving the saved entries
    # Note that the change may result in a modified list order, so callers should AWAYS
    # use the returned list or get a new one as that will reflect the current state of
    # afairs
    def add_entry(self, entry):
        if entry.start_time > entry.stop_time:
            # this is a over midnight situation
            self._add_entry(TimeofUse(entry.start_time, time.max, entry.max_power))
            self._add_entry(TimeofUse(time.min, entry.stop_time,entry.max_power))
        else:
            self._add_entry(entry)
        return self.get_as_tou_schedule()

    # Note that the change may result in a modified list order, so callers should AWAYS
    # use the returned list or get a new one as that will reflect the current state of
    # afairs
    def delete_entry(self, entry_number):
        self._schedule_entries.pop(entry_number)
        return self.get_as_tou_schedule()
    # removes and exisitng entry and adds a new one , this is really just a convenience
    # If the new entry is rejected due to overlap then the deleted one IS NOT REPLACED
    # Note that the change may result in a modified list order, so callers should AWAYS
    # use the returned list or get a new one as that will reflect the current state of
    # afairs
    def remove_and_replace_entry(self, old_entry_nuber, new_entry):
        self._schedule_entries.pop(old_entry_nuber)
        return self.add_entry(new_entry)

    def get_as_tou_schedule(self)-> List[TimeofUse]:
        schedules = []
        for i in self._schedule_entries :
            schedules.append(i.get_as_tou())
        return schedules
    
    def get_as_string(self) -> str:
        result = ""
        done_first = False
        for entry in self._schedule_entries :
            if done_first:
                result = result +","
            else :
                done_first = True
            result = result + str(entry.get_as_string())
        return result 
    
    # retained fore compatibility purposed, is not just a wrapper around `def load_tou_schedule`
    def load_tou_schedule(self, schedule):
        self.load_tou_schedule_from_json(schedule)

    # replace the current tou schedule data with the new dicitonary data
    def load_tou_schedule_from_json(self, json_schedule):
        self._schedule_entries = []
        for entry in json_schedule:
            tou_entry = create_time_of_use_from_tou(entry)
            self.add_entry(tou_entry)

    # create a new timeofuseschedule with the provided array ofdictionary data

    def entry_count(self) -> int:
        return len(self._schedule_entries)

    
    def get_tou_entry_count(self) -> int:
        return len(self._schedule_entries)
    
    def get_tou_entry(self, i:int) -> TimeofUse|None:
        entrties = self.get_tou_entry_count()
        if i > entrties:
            return None
        else:
            return self._schedule_entries[i]

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        my_entry_count = self.entry_count() 
        other_entry_count = other.entry_count()
        # if there both zero length by definition they are equal
        if (my_entry_count == 0) and (other_entry_count==0):
            return True
        # different numbers of entries means different scheduled
        if my_entry_count != other_entry_count:
            return False
        # for each entry
        for i in range(0, my_entry_count):
            my_tou = self.get_tou_entry(i)
            other_tou = other.get_tou_entry(i)
            if my_tou != other_tou:
                return False
        # got to the end of the individual timeofuse entries and they arew all equal so ...
        return True
    
    def __hash__(self) -> int:
        my_hash = 0
        my_entry_count = self.entry_count() 
        for i in range(0, my_entry_count):
            my_tou = self.get_tou_entry(i)
            my_tou_hash = hash(my_tou)
            # adjust the hash based on the position in the order to allow for things in a differing order
            my_hash = my_hash ^ (my_tou_hash + i)
        return my_hash



#!/usr/bin/env python3
# I suspect that I haven't got to grips with the way phthon does things, but soppusely this will setup the path to allod for the sonnen batteri moduel to be in a separate location
# To me having to do this for testing seems a horrendous hack
import os
import sys
import time
script_path = os.path.realpath(os.path.dirname(__name__))
os.chdir(script_path)
sys.path.append("..")

from login import *
from pprint import pprint
from sonnenbatterie2 import SonnenBatterieV2
from timeofuse import TimeofUseSchedule
from timeofuse.timeofuse import create_time_of_use_entry
from sonnenbatterie2.const import SONNEN_OPERATING_MODE_TIME_OF_USE_NAME

def test_timeofuse():
    sb = SonnenBatterieV2(SONNEN_IP, SONNEN_TOKEN)
    operating_mode_name = sb.get_operating_mode_name()
    print(f"\nOriginal operating mode is {operating_mode_name}")
    print(f"\nSetting operating mode name to {SONNEN_OPERATING_MODE_TIME_OF_USE_NAME}")
    pprint(sb.set_operating_mode_by_name(SONNEN_OPERATING_MODE_TIME_OF_USE_NAME))
    print(f"\nSet operating mode to {sb.get_operating_mode_name()}")
    tous = TimeofUseSchedule()
    print("\nExtract tou schedule as string")
    orig_battery_tou_string=sb.get_tou_schedule_string()
    orig_battery_tou_json = sb.get_tou_schedule_json()
    pprint(sb.get_tou_schedule_string())
    print("\nExtract tou schedule as json objects")
    pprint(sb.get_tou_schedule_json())
    print("\nExtract tou schedule as text schedule")
    pprint(sb.get_tou_schedule_object())
    print("\nLoad tou schedule from json")
    tous.load_tou_schedule_from_json(sb.get_tou_schedule_json())
    print("\nLoaded tou schedule")
    print(tous.get_as_tou_schedule())
    ### Disabled since Sonnen seems to have changed the format
    print ("\nCreate a new TOU Schedule 10:00 - 11:00 and 14:00 - 15:00")
    tous_new = TimeofUseSchedule()
    tous_new.add_entry(create_time_of_use_entry(10,00,11,00))
    tous_new.add_entry(create_time_of_use_entry(14,00,15,00))
    print("Setting new TOU schedule to ")
    print(tous_new.get_as_tou_schedule())
    sb.set_tou_schedule_json(tous_new.get_as_tou_schedule())
    print("Sleeping for 60 seconds so you can check the change has been applied")
    time.sleep(60)
    print("\nExtract replaced tou schedule as objects")
    pprint(sb.get_tou_schedule_json())
    print("\nExtract restored tou schedule as string")
    pprint(sb.get_tou_schedule_string())
    ### Disabled since Sonnen seems to have changed the format
    print(f"\nRestoring origional TOU schedule to {orig_battery_tou_string}")
    sb.set_tou_schedule_json(orig_battery_tou_json)
    print("Sleeping for 60 seconds so you can check the change has been applied")
    time.sleep(60)
    print(f"\nResetting operating mode name to {operating_mode_name}")
    pprint(sb.set_operating_mode_by_name(operating_mode_name))
    print(f"\nReset operating mode to {sb.get_operating_mode_name()}")

if __name__ == '__main__':
    test_timeofuse()


DEFAULT_BATTERY_LOGIN_TIMEOUT=10
DEFAULT_CONNECT_TO_BATTERY_TIMEOUT=6
DEFAULT_READ_FROM_BATTERY_TIMEOUT=6

SONNEN_OPERATING_MODE_UNKNOWN_NAME="Unknown Operating Mode"
SONNEN_OPERATING_MODE_UNKNOWN="0"
SONNEN_OPERATING_MODE_MANUAL_NAME="Manual"
SONNEN_OPERATING_MODE_AUTOMATIC_SELF_CONSUMPTION_NAME="Automatic - Self-Consumption"
SONNEN_OPERATING_MODE_BATTERY_MODULE_EXTENSION_30_PERCENT_NAME="Battery-Module-Extension (30%)"
SONNEN_OPERATING_MODE_TIME_OF_USE_NAME="Time-Of-Use"
SONNEN_OPERATING_MODE_NAMES_TO_OPERATING_MODES =  {
    SONNEN_OPERATING_MODE_MANUAL_NAME : 1,
    SONNEN_OPERATING_MODE_AUTOMATIC_SELF_CONSUMPTION_NAME: 2,
    SONNEN_OPERATING_MODE_BATTERY_MODULE_EXTENSION_30_PERCENT_NAME: 6,
    SONNEN_OPERATING_MODE_TIME_OF_USE_NAME: 10
}
SONNEN_OPERATING_MODES_TO_OPERATING_MODE_NAMES = {v:k for k,v in SONNEN_OPERATING_MODE_NAMES_TO_OPERATING_MODES.items()}

SONNEN_OPEATING_MODES=[SONNEN_OPERATING_MODE_MANUAL_NAME, SONNEN_OPERATING_MODE_AUTOMATIC_SELF_CONSUMPTION_NAME, SONNEN_OPERATING_MODE_BATTERY_MODULE_EXTENSION_30_PERCENT_NAME,SONNEN_OPERATING_MODE_TIME_OF_USE_NAME ]

SONNEN_CONFIGURATION_OPERATING_MODE="EM_OperatingMode"
SONNEN_CONFIGURATION_TOU_SCHEDULE="EM_ToU_Schedule"
SONNEN_CONFIGURATION_BACKUP_RESERVE="EM_USOC"
SONNEN_LATEST_DATA_CHARGE_LEVEL="USOC"

SONNEN_API_PATH_CONFIGURATIONS="v2/configurations"
SONNEN_API_PATH_LATEST_DATA="v2/latestdata"
SONNEN_API_PATH_POWER_METER="powermeter"
SONNEN_API_PATH_BATTERY_SYSTEM="battery_system"
SONNEN_API_PATH_INVERTER="inverter"
SONNEN_API_PATH_SYSTEM_DATA="system_data"
SONNEN_API_PATH_STATUS="v1/status"
SONNEN_API_PATH_BATTERY="battery"

SONNEN_CHARGE_PATH="charge"
SONNEN_DISCHARGE_PATH="discharge"



#!/usr/bin/env python3
# I suspect that I haven't got to grips with the way phthon does things, but soppusely this will setup the path to allod for the sonnen batteri moduel to be in a separate location
# To me having to do this for testing seems a horrendous hack
import json
import os
import sys

script_path = os.path.realpath(os.path.dirname(__name__))
os.chdir(script_path)
sys.path.append("..")
from pprint import pprint
from timeofuse.timeofuse import TimeofUseSchedule, create_time_of_use_entry

problemHit = False
problemsList = ""
def do_test(description:str, test, expected_result):
    actual_result = test()
    if expected_result == actual_result:
        print (description+" achieved expected result of "+str(actual_result))
    else :
        global problemHit
        global problemsList
        problemHit = True
        problemstr = description+" did not result in expected result of "+str(expected_result)+" but actually resulted in "+str(actual_result)
        print ("================= PROBLEM ===============")
        print (problemstr)
        print ("================= PROBLEM ===============")
        problemsList = problemsList + "\n"+problemstr

if __name__ == '__main__':
    tous = TimeofUseSchedule()  
    print("\nEmpty schedule")
    pprint(tous.get_as_tou_schedule())
    tou=create_time_of_use_entry()
    print("\nOverlapping midnight schedule")
    pprint(tous.add_entry(tou))
    print ("\nAdding non overlaping entry")
    tou=create_time_of_use_entry(10,0,11,0)
    touTenToEleven = tou
    pprint(tous.add_entry(tou))
    print ("\nAdding fully overlaping entry")
    tou=create_time_of_use_entry(9,0,12,0)
    touNineToMidday = tou
    try:
        tous.add_entry(tou)
        print ("\nOpps, no exception, tnis is a bug")
    except Exception as e:
        print ("\nExpected exception message "+str(e.args))

    pprint(tous.get_as_tou_schedule())

    print("Getting as string")
    print(tous.get_as_string())

    print ("\nAdding overlaping start entry")
    tou=create_time_of_use_entry(9,0,10,30)
    touNineToTenThirty = tou
    try:
        tous.add_entry(tou)
        print ("\nOpps, no exception, tnis is a bug")
    except Exception as e:
        print ("\nExpected exception message "+str(e.args))
    pprint(tous.get_as_tou_schedule())

    print ("\nAdding overlaping end entry")
    tou=create_time_of_use_entry(10,30, 11, 30)
    try:
        tous.add_entry(tou)
        print ("\nOpps, no exception, tnis is a bug")
    except Exception as e:
        print ("\nExpected exception message "+str(e.args))
    pprint(tous.get_as_tou_schedule())

    print ("\nAdding exact match overlaping entry")
    tou=create_time_of_use_entry(10,0,11,0)
    try:
        tous.add_entry(tou)
        print ("\nOpps, no exception, this is a bug")
    except Exception as e:
        print ("\nExpected exception message "+str(e.args))
    pprint(tous.get_as_tou_schedule())

    print ("\nBuilding based on returned entry")
    old_schedule = tous.get_as_tou_schedule()
    tous = TimeofUseSchedule() 
    tous.load_tou_schedule(old_schedule)

    print("\nAfter schedule load")
    pprint(tous.get_as_tou_schedule())
    print("\nAfter deleting index entry")
    pprint(tous.delete_entry(1))
    js = json.dumps(tous.get_as_tou_schedule())
    print("Dumped json data "+js)

    print("Testing timeofuse class hash codes")
    touSecondNineToMidday = create_time_of_use_entry(9,0,12,0)

    touTouTenToElevenMaxPower=create_time_of_use_entry(10,0,11,0, 40000)
    firstNineToMiddayHash = hash(touNineToMidday)
    secondNineToMiddayHash = hash(touSecondNineToMidday)
    tenToElevenHash = hash(touTenToEleven)
    touTouTenToElevenMaxPowerHash = hash (touTouTenToElevenMaxPower)
    print("Hash of first nine to midday "+str(firstNineToMiddayHash))
    print("Hash of second nine to midday "+str(secondNineToMiddayHash))
    print("Hash of ten to eleven "+str(tenToElevenHash))
    print("Hash of ten to eleven max power "+str(touTouTenToElevenMaxPowerHash))
    do_test("Compare hash of first and second nine to midday  ", lambda:firstNineToMiddayHash == secondNineToMiddayHash, True)
    do_test("Compare hash of nine to midday and ten to eleven ", lambda:firstNineToMiddayHash == tenToElevenHash, False)
    do_test("Compare hash of ten to eleven and ten to eleven max power ", lambda:touTouTenToElevenMaxPowerHash == tenToElevenHash, False)

    print("Testing timeofuse class equality")
    touSecondNineToMidday = create_time_of_use_entry(9,0,12,0)
    do_test("Compare first nine to midday with itself ", lambda:touNineToMidday == touNineToMidday, True)
    do_test("Compare first nine to midday with sesond nine to midday ", lambda:touNineToMidday == touSecondNineToMidday, True)
    do_test("Compare second nine to midday with first nine to midday ", lambda:touSecondNineToMidday == touNineToMidday, True)
    do_test("Compare nine to midday with ten to eleven ", lambda:touNineToMidday == touTenToEleven, False)
    do_test("Compare ten to eleven with ten to eleven max power ", lambda: touTenToEleven == touTouTenToElevenMaxPower , False)

    
    firstEmptyTous = TimeofUseSchedule()
    secondEmptyTous = TimeofUseSchedule()

    firstNineToMiddayTous = TimeofUseSchedule()
    firstNineToMiddayTous.add_entry(touNineToMidday)
    firstNineToMiddayTousSameTou = TimeofUseSchedule()
    firstNineToMiddayTousSameTou.add_entry(touNineToMidday)
    secondNineToMiddayTous = TimeofUseSchedule()
    secondNineToMiddayTous.add_entry(touSecondNineToMidday)
    tenToElevenTous = TimeofUseSchedule()
    tenToElevenTous.add_entry(touTenToEleven)

    print("Testing single entry time of use schedule hashes")
    firstEmptyTousHash = hash(firstEmptyTous)
    secondEmptyTousHash = hash(secondEmptyTous)
    firstNineToMiddayTousHash = hash(firstNineToMiddayTous)
    firstNineToMiddayTousSameTouHash = hash(firstNineToMiddayTousSameTou)
    secondNineToMiddayTousHash = hash(secondNineToMiddayTous)
    tenToElevelTousHash = hash(tenToElevenTous)
    print("Hash of first empty tous "+str(firstEmptyTousHash))
    print("Hash of second empty tous "+str(secondEmptyTousHash))
    print("Hash of first nine to midday tous "+str(firstNineToMiddayTousHash))
    print("Hash of first nine to midday tous with same tou "+str(firstNineToMiddayTousSameTouHash))
    print("Hash of second nine to midday tous "+str(secondNineToMiddayTousHash))
    print("Hash of ten to eleven  tous "+str(tenToElevelTousHash))
    do_test("Compare hash of first empty with itself ", lambda:firstEmptyTousHash == firstEmptyTousHash,True)
    do_test("Compare hash of first and second empty ", lambda:firstEmptyTousHash == secondEmptyTousHash,True)
    do_test("Compare hash of empty and nine to midday ", lambda:firstEmptyTousHash == firstNineToMiddayTousHash,False)
    do_test("Compare hash of first midday with itself ", lambda:firstNineToMiddayTousHash == firstNineToMiddayTousHash,True)
    do_test("Compare hash of first midday with other nine to middat using tame tou  ", lambda:firstNineToMiddayTousHash == firstNineToMiddayTousSameTouHash,True)
    do_test("Compare hash of first and second nine to midday ", lambda:firstNineToMiddayTousHash == secondNineToMiddayTousHash,True)
    do_test("Compare hash of first nine to midday with ten to elevel ", lambda:firstNineToMiddayTousHash == tenToElevelTousHash,False)

    print("Testing single entry time of use schedule class equality")

    do_test("Compare first empty tous with itself ", lambda:firstEmptyTous == firstEmptyTous,True)
    do_test("Compare first empty tous with second empty ", lambda:firstEmptyTous == secondEmptyTous,True)
    do_test("Compare second empty tous with first empty ", lambda:secondEmptyTous==firstEmptyTous,True)
    do_test("Compare first nine to midday tous with itself ", lambda:firstNineToMiddayTous == firstNineToMiddayTous,True)
    do_test("Compare first nine to midday tous with the same using the same Tou ", lambda:firstNineToMiddayTous == firstNineToMiddayTousSameTou,True)
    do_test("Compare first empty tous with first nine to midday ", lambda:firstEmptyTous == firstNineToMiddayTous,False)
    do_test("Compare first nine to midday tous with first empty ", lambda:firstNineToMiddayTous==firstEmptyTous,False)
    do_test("Compare first nine to midday tous with ten to eleven  ", lambda:firstNineToMiddayTous==tenToElevenTous,False)
    do_test("Compare ten to eleven tous with first nine to midday  ", lambda:tenToElevenTous==firstNineToMiddayTous,False)


    print("Testing multiple entry time of use schedule hashes")
    morningTou = create_time_of_use_entry(9,0,10,0)
    afternoonTou = create_time_of_use_entry(13,0,15,0)
    eveningTou = create_time_of_use_entry(21,0,22,0)
    nightTou = create_time_of_use_entry(1,30,3,40)
    overMidnightTou = create_time_of_use_entry(23,30,5,30)
    uptoMidnightTou = create_time_of_use_entry(23,30,0,0)
    afterMidnightTou = create_time_of_use_entry(0,0,5, 30)

    tousMorningAfternoon = TimeofUseSchedule() 
    tousMorningAfternoon.add_entry(morningTou)
    tousMorningAfternoon.add_entry(afternoonTou)
    tousFirstMorningAfternoonEveningNight = TimeofUseSchedule() 
    tousFirstMorningAfternoonEveningNight.add_entry(morningTou)
    tousFirstMorningAfternoonEveningNight.add_entry(afternoonTou)
    tousFirstMorningAfternoonEveningNight.add_entry(eveningTou)
    tousFirstMorningAfternoonEveningNight.add_entry(nightTou)
    tousSecondMorningAfternoonEveningNight = TimeofUseSchedule() 
    tousSecondMorningAfternoonEveningNight.add_entry(morningTou)
    tousSecondMorningAfternoonEveningNight.add_entry(afternoonTou)
    tousSecondMorningAfternoonEveningNight.add_entry(eveningTou)
    tousSecondMorningAfternoonEveningNight.add_entry(nightTou)
    tousAfternoonMorning = TimeofUseSchedule() 
    tousAfternoonMorning.add_entry(afternoonTou)
    tousAfternoonMorning.add_entry(morningTou)
    tousAfternoonEvening = TimeofUseSchedule() 
    tousAfternoonEvening.add_entry(afternoonTou)
    tousAfternoonEvening.add_entry(eveningTou)
    tousOvernight = TimeofUseSchedule() 
    tousOvernight.add_entry(overMidnightTou)
    tousUptoAfterMidnight =  TimeofUseSchedule() 
    tousUptoAfterMidnight.add_entry(uptoMidnightTou)
    tousUptoAfterMidnight.add_entry(afterMidnightTou)

    do_test("Compare hashes for first and second  morning/afternoon/evening/night", lambda:hash(tousFirstMorningAfternoonEveningNight) == hash (tousSecondMorningAfternoonEveningNight), True)
    do_test("Compare hashes for morning / afternoon with first  morning/afternoon/evening/night", lambda:hash(tousMorningAfternoon) == hash (tousSecondMorningAfternoonEveningNight), False)
    do_test("Compare hashes for morning/afternoon with afternoon/morning", lambda:hash(tousMorningAfternoon) == hash (tousAfternoonMorning), True)
    do_test("Compare hashes for  afternoon/morning with morning/afternoon", lambda:hash(tousAfternoonMorning) == hash (tousMorningAfternoon), True)
    do_test("Compare hashes for upto after midnight with overnight", lambda:hash(tousUptoAfterMidnight) == hash (tousOvernight), True)


    print("Testing multiple entry time of use schedule class equality")
    do_test("Comparing single and multiple entry tous (morning / afternoon / evening / night with nine to midday )", lambda: tousFirstMorningAfternoonEveningNight == firstNineToMiddayTous, False )
    do_test("Comparing multiple and single entry tous (nine to midday with morning / afternoon / evening / night  )", lambda: firstNineToMiddayTous == tousFirstMorningAfternoonEveningNight, False )
    do_test("Comparing  morning/afternoon with afternoon/morning )", lambda: tousMorningAfternoon == tousAfternoonMorning, True )
    do_test("Comparing afternoon/morning with   morning/afternoon )", lambda: tousAfternoonMorning == tousMorningAfternoon, True )
    do_test("Comparing overnight with upto / after midnight )", lambda: tousOvernight == tousUptoAfterMidnight, True )
    do_test("Comparing  upto / after midnight with overnight)", lambda: tousUptoAfterMidnight==tousOvernight, True )


    if problemHit:
        print ("=========== WARNING, WARNING, WARNING ===========")
        print (problemsList)
        print ("=========== WARNING, WARNING, WARNING ===========")
    else:
        print("All worked fine")



#!/usr/bin/env python3
# I suspect that I haven't got to grips with the way phthon does things, but soppusely this will setup the path to allod for the sonnen batteri moduel to be in a separate location
# To me having to do this for testing seems a horrendous hack
import os
import sys
import time

import requests

from login import *

script_path = os.path.realpath(os.path.dirname(__name__))
os.chdir(script_path)
sys.path.append("..")
from sonnenbatterie.sonnenbatterie import sonnenbatterie

def main():
    print("Starting login")
    try:
        sb = sonnenbatterie(SONNEN_USERNAME, SONNEN_PASSWORD, SONNEN_IP)
    except requests.exceptions.Timeout as e:
        print("Timeout connection "+str(type(e))+", details "+str(e))    
        exit()
    except Exception as e:
        print("non timeout exception getting connection"+str(type(e))+", details "+str(e)) 
        exit()
    print("Logged in")
    sb.set_request_connect_timeout(30)
    sb.set_request_read_timeout(30)
    counter = 0
    while True and counter < 5:
        counter = counter + 1
        reserve = -1
        current_level = -1
        tou = "Not retrieved"
        mode = "Not retrieved"
        status = "Not Retrieved"
        try:
            reserve = sb.get_battery_reserve()
        except requests.exceptions.Timeout as e:
            print("Timeout getting reserve "+str(type(e))+", details "+str(e))    
        except Exception as e:
            print("non timeout exception getting reserve "+str(type(e))+", details "+str(e)) 
        
        try:
            tou = sb.sb2.get_tou_schedule_string()
        except requests.exceptions.Timeout as e:
            print("Timeout getting tou "+str(type(e))+", details "+str(e))    
        except Exception as e:
            print("non timeout exception getting tou "+str(type(e))+", details "+str(e)) 

        try:
            mode = sb.get_operating_mode_name()
        except requests.exceptions.Timeout as e:
            print("Timeout getting mode "+str(type(e))+", details "+str(e))    
        except Exception as e:
            print("non timeout exception getting mode "+str(type(e))+", details "+str(e)) 

        try:
            status = sb.get_status()
            current_level = status["RemainingCapacity_Wh"]
        except requests.exceptions.Timeout as e:
            print("Timeout getting tou "+str(type(e))+", details "+str(e))    
        except Exception as e:
            print("non timeout exception getting tou "+str(type(e))+", details "+str(e)) 

        print("Loop "+str(counter)+", retrireved data : current_level "+str(current_level)+", reserve "+str(reserve)+", mode "+mode+", tou "+tou)
        time.sleep(15)

if __name__ == '__main__':
    main()


#!/usr/bin/env python3
# I suspect that I haven't got to grips with the way phthon does things, but soppusely this will setup the path to allod for the sonnen batteri moduel to be in a separate location
# To me having to do this for testing seems a horrendous hack
import asyncio
import os
import sys
import time

script_path = os.path.realpath(os.path.dirname(__name__))
os.chdir(script_path)
sys.path.append("..")
from login import *
from pprint import pprint
from sonnenbatterie.sonnenbatterie import sonnenbatterie, AsyncSonnenBatterie


# this is based on the test code by rust dust

def main():
  print("\nMain (sync)\n==========")
  sb = sonnenbatterie(SONNEN_USERNAME, SONNEN_PASSWORD, SONNEN_IP)
  print("\nPower:\n")
  pprint(sb.get_powermeter())
  print("\nBattery System:\n")
  pprint(sb.get_batterysystem())
  print("\nInverter:\n")
  pprint(sb.get_inverter())
  print("\nSystem Data:\n")
  pprint(sb.get_systemdata())
  print("\nStatus:\n")
  pprint(sb.get_status())
  print("\nBattery:\n")
  pprint(sb.get_battery())

async def async_main():
  print("\n\nAsync main\n==========\n")
  sb = AsyncSonnenBatterie(SONNEN_USERNAME, SONNEN_PASSWORD, SONNEN_IP)
  print("\nPower:\n")
  pprint(await sb.get_powermeter())
  print("\nBattery System:\n")
  pprint(await sb.get_batterysystem())
  print("\nInverter:\n")
  pprint(await sb.get_inverter())
  print("\nSystem Data:\n")
  pprint(await sb.get_systemdata())
  print("\nStatus:\n")
  pprint(await sb.get_status())
  print("\nBattery:\n")
  pprint(await sb.get_battery())
  await sb.logout()

if __name__ == '__main__':
  start = time.time()
  main()
  print("\nTotal time sync:", time.time() - start)
  start = time.time()
  asyncio.run(async_main())
  print("\nTotal time async:", time.time() - start)



#!/usr/bin/env python3
# I suspect that I haven't got to grips with the way phthon does things, but soppusely this will setup the path to allod for the sonnen batteri moduel to be in a separate location
# To me having to do this for testing seems a horrendous hack
import asyncio
import os
import sys
script_path = os.path.realpath(os.path.dirname(__name__))
os.chdir(script_path)
sys.path.append("..")

from login import *
from pprint import pprint
from sonnenbatterie import AsyncSonnenBatterie
from sonnenbatterie.const import *

async def main():
  sb = AsyncSonnenBatterie(SONNEN_USERNAME, SONNEN_PASSWORD, SONNEN_IP)
  operating_mode_num = await sb.get_operating_mode()
  operating_mode_name = await sb.get_operating_mode_name()
  print("\nConfigurations")
  pprint(await sb.get_configurations())
  print("\nOperating mode (num)")
  pprint(await sb.get_operating_mode())
  print("\nOperating mode (name)")
  pprint(await sb.get_operating_mode_name())
  print("Setting operating mode num to 1")
  pprint(await sb.set_operating_mode(1))
  print("\nNew Operating mode (num)")
  pprint(await sb.get_operating_mode())
  print("\nNew Operating mode (name)")
  pprint(await sb.get_operating_mode_name())
  print(f"Resetting operating mode num to {operating_mode_num}")
  pprint(await sb.set_operating_mode(operating_mode_num))
  print("\nReset Operating mode (num)")
  pprint(await sb.get_operating_mode())
  print("\nReset Operating mode (name)")
  pprint(await sb.get_operating_mode_name())
  print("Setting operating mode name to "+SONNEN_OPERATING_MODE_AUTOMATIC_SELF_CONSUMPTION_NAME)
  pprint(await sb.set_operating_mode_by_name(SONNEN_OPERATING_MODE_AUTOMATIC_SELF_CONSUMPTION_NAME))
  print("\nNew Operating mode (num)")
  pprint(await sb.get_operating_mode())
  print("\nNew Operating mode (name)")
  pprint(await sb.get_operating_mode_name())
  print(f"Resetting operating mode num to {operating_mode_num}")
  pprint(await sb.set_operating_mode(operating_mode_num))
  print("\nVerifiying operating mode")
  pprint(await sb.get_operating_mode())
  await sb.logout()

if __name__ == '__main__':
  asyncio.run(main())
